import logging
import sys
import importlib
import threading 

from pathlib import Path

from . import logger as pyrsched_logger

logger = pyrsched_logger.getChild('import')

def resolve_module_file(file_name):
    module_file = Path(file_name)
    if not module_file.is_absolute():
        module_file = Path(__file__).parent.parent.parent / module_file
    if not module_file.exists():
        raise FileNotFoundError(f"{module_file} not found.")
    return module_file

def check_search_path(module_file):
    module_path = module_file.parent.resolve()
    if not str(module_path) in sys.path:
        logger.info(f'{module_path} not in sys.path, adding it')
        sys.path.insert(0, str(module_path))

def import_external_attribute(file_name, attribute_name):  # pragma: no cover; this is not used
    imported_module = import_external(file_name)
    logger.info(f'importing {attribute_name} from {imported_module}')
    attribute = getattr(imported_module, attribute_name)
    return attribute

def import_external(file_name):
    logger.info(f'importing {file_name}')
    module_file = resolve_module_file(file_name)
    check_search_path(module_file)
    imported_module = importlib.import_module(module_file.stem)
    return imported_module



import logging
import logging.config
import os
import threading
import uuid
from copy import deepcopy
from multiprocessing.managers import BaseManager
from pathlib import Path

from apscheduler.schedulers.background import BackgroundScheduler

from . import logger
from .service import SchedulerService
from .utils import import_external

class ServerWrapper(object):
    def __init__(self, conf_dir):
        self._conf_dir = conf_dir
        self._config_file = Path(self._conf_dir) / "scheduler_config.py"
        self._logger = logger.getChild("ServerWrapper")

        self._config = import_external(self._config_file)
        logging.config.dictConfig(self._config.log_config)

    def _check_authkey(self):
        # generate an authkey if it was not provided via environment
        authkey = os.environ.get("PYRSCHED_SECRET", None)
        if not authkey:
            self._logger.warning(
                "WARNING! Shared secret not set. Use the following generated secret for the clients:"
            )
            authkey = uuid.uuid4()
        else:
            self._logger.warning("WARNING! Using shared secret from the environment:")
        self._logger.warning(authkey)
        return authkey

    def shutdown(self):  # pragma: no cover
        # self._server.shutdown()
        try:
            self._scheduler.shutdown()
        except:
            pass

    def start(self):  # pragma: no cover
        if self._logger.level <= logging.DEBUG:
            self._logger.debug("Starting server with config:")
            for k, v in self._config.pypyr.items():
                self._logger.debug(f"{k}: {v}")

        try:
            scheduler_config = deepcopy(self._config.apscheduler)
            scheduler = BackgroundScheduler(scheduler_config)

            class SchedulerManager(BaseManager):
                pass

            authkey = self._check_authkey()
            service = SchedulerService(
                scheduler=scheduler, config=self._config, logger=self._logger
            )
            SchedulerManager.register("scheduler", callable=lambda: service)

            manager = SchedulerManager(
                address=("", self._config.pypyr.get("server_port", 12345)),
                authkey=str(authkey).encode("utf-8"),
            )
            server = manager.get_server()

            scheduler.start()
            server.serve_forever()

        except OSError as ose:
            self._logger.error(ose.strerror)
            


import threading
import logging
import re


class PipelineLoggingContext:
    def __init__(self, logger, loglevel=None, handler=None, filter_list=None):
        self.logger = logger
        self.handler = handler
        self.filter_list = filter_list
        self.loglevel = loglevel
        self.lock = threading.Lock()

    def __enter__(self):
        self.lock.acquire()
        if self.loglevel is not None:
            self._old_level = self.logger.level
            self.logger.setLevel(self.loglevel)
        if self.handler:
            for f in self.filter_list:
                self.handler.addFilter(f)
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        self.lock.release()
        if self.loglevel is not None:
            self.logger.setLevel(self._old_level)
        if self.handler:
            for f in self.filter_list:
                self.handler.removeFilter(f)
            self.logger.removeHandler(self.handler)


class SensitiveValueFilter(logging.Filter):
    """
    Logging filter which censors sensitive information.

    The fitler checks if the log message containssensitive key-value-pairs.
    These pairs have to occure in the message in the form "'<key>': '<value>'". 
    This does not have to be a string representation of a dictionary, but will be in most cases.

    Then the value will be replaced with the string '*****' and the log message will be emitted
    by its handler.
    
    """
    def __init__(self, name='', sensitive_keys=None):
        super().__init__(name)
        self.sensitive_keys = sensitive_keys if sensitive_keys else []
        self.regex_list = []
        for key in self.sensitive_keys:            
            self.regex_list.append( re.compile(r"\'(?P<key>" + key + r")\':\s*\'.*?\'") )

    def filter(self, record):
        for r in self.regex_list:
           record.msg = r.sub("'\g<key>': '*****'", record.msg)
        return True



import uuid
import logging
from pathlib import Path

import psutil

from apscheduler.triggers.interval import IntervalTrigger
from apscheduler.jobstores.base import JobLookupError
from pypyr.pipelinerunner import main as pipeline_runner

from .logging import PipelineLoggingContext, SensitiveValueFilter

NEW_JOB_MAX_INSTANCES = 1

def job_function(pipeline_name, log_path, log_level, log_format, pipeline_path, sensitive_keys):
    logger = logging.getLogger("pypyr")
    log_filename = Path(log_path) / f"{pipeline_name}.log"

    # ToDo: could the handler be instantiated outside and be reused?
    pipeline_handler = logging.FileHandler(log_filename)
    pipeline_handler.setFormatter(logging.Formatter(fmt=log_format))

    log_filter = SensitiveValueFilter(sensitive_keys=sensitive_keys)
    with PipelineLoggingContext(logger, loglevel=log_level, handler=pipeline_handler, filter_list=[log_filter, ]):
        pipeline_runner(
            pipeline_name, pipeline_context_input="", working_dir=Path(pipeline_path),
        )


class SchedulerService(object):
    def __init__(self, scheduler=None, config=None, logger=None):
        self._logger = logger.getChild("service") if logger else logging.getLogger(
            "pyrsched.service"
        )
        self._config = config
        self._scheduler = scheduler

    def _marshal_job(self, job):
        """ make a data structure which is able to be submitted over the RPC line"""
        marshalled_job = job.__getstate__()
        marshalled_job["next_run_time"] = (
            marshalled_job["next_run_time"].isoformat()
            if marshalled_job["next_run_time"]
            else None
        )

        # add an "is_running" flag
        marshalled_job["is_running"] = marshalled_job["next_run_time"] is not None
        trigger = job.trigger.__getstate__()
        marshalled_job["trigger"] = {
            "interval": trigger["interval"].total_seconds(),
            "start_date": trigger["start_date"].isoformat()
            if trigger["start_date"]
            else None,
            "timezone": str(trigger["timezone"]),
        }

        return marshalled_job

    def add_job(self, pipeline_name, interval=60):
        self._logger.info(f"add_job({pipeline_name}, {interval})")

        # ToDo: lookup if a job with the same name already exists and handle duplicate job names (suffix?)

        formatters = self._config.log_config.get(
            "formatters",
            {
                "standard": {
                    "format": "%(asctime)s [%(levelname)s] %(name)s: %(message)s"
                }
            },
        )
        log_format = formatters["standard"]["format"]
        log_path = str(
            Path(self._config.pypyr.get("pipelines.log_path", Path("logs"))).resolve()
        )
        pipeline_path = str(
            Path(self._config.pypyr.get("pipelines.base_path", Path("pipelines"))).resolve()
        )

        job = self._scheduler.add_job(
            job_function,
            id=str(uuid.uuid4()),
            name=pipeline_name,
            trigger=IntervalTrigger(seconds=interval),
            next_run_time=None,
            args=[pipeline_name,],
            max_instances=NEW_JOB_MAX_INSTANCES,
            misfire_grace_time=None,
            coalesce=self._config.apscheduler.get("apscheduler.job_defaults.coalesce", False),
            kwargs={
                "log_path": log_path,
                "log_format": log_format,
                "log_level": self._config.pypyr.get(
                    "pipelines.log_level", logging.WARNING
                ),
                "pipeline_path": pipeline_path, 
                "sensitive_keys": self._config.pypyr.get(
                    "pipelines.sensitive_keywords", None
                ),
            },
        )

        return job.id

    def reschedule_job(self, job_id, interval=60):
        self._logger.info(f"reschedule_job({job_id})")
        trigger = IntervalTrigger(seconds=interval)
        try:
            job = self._scheduler.reschedule_job(job_id, trigger=trigger)
        except JobLookupError as jle:
            self._logger.exception(jle, exc_info=False)
            return None
        return self._marshal_job(job)

    def pause_job(self, job_id):
        self._logger.info(f"pause_job({job_id})")
        try:
            job = self._scheduler.pause_job(job_id)
        except JobLookupError as jle:
            self._logger.exception(jle, exc_info=False)
            return None
        return self._marshal_job(job)

    def start_job(self, job_id):
        """ Start a job. 

        To start a job, it has to be present in the jobstore, otherwise this method
        will return ``None`` over the wire. 

        If the job exists in the jobstore, the scheduler is advised to start the job.

        :param job_id: The ID of the job to be started
        :type job_id: string
        :return: the job which was started or None if the job was not found in the jobstore
        :rtype: Job state or None
        """
        self._logger.info(f"start_job({job_id})")
        try:
            job = self._scheduler.resume_job(job_id)
        except JobLookupError as jle:
            self._logger.exception(jle, exc_info=False)
            return None
        return self._marshal_job(job)

    def remove_job(self, job_id):
        self._logger.info(f"remove_job({job_id})")
        try:
            self._scheduler.remove_job(job_id)
        finally:
            return None  # self._marshal_job(job_id)

    def get_job(self, job_id):
        self._logger.info(f"get_job({job_id})")
        job = self._scheduler.get_job(job_id)
        return self._marshal_job(job) if job is not None else None

    def list_jobs(self):
        self._logger.info(f"list_jobs()")
        job_list = self._scheduler.get_jobs()
        return [self._marshal_job(job) for job in job_list]

    def state(self):
        self._logger.info("state()")
        job_list = self._scheduler.get_jobs()
        state_obj = {
            "run_state": self._scheduler.state,
            "is_running": self._scheduler.running,
            "total_jobs": len(job_list),
            "job_list": [self._marshal_job(job) for job in job_list],
            "cpu_load": psutil.getloadavg(),
        }
        return state_obj



from pathlib import Path

import click

from . import logger
from .server import ServerWrapper


@click.command()
@click.option("-c", "--conf", "conf_dir", default="conf", help="config directory")
def cli(conf_dir):
    click.echo(f"Starting scheduler. Config: {Path(conf_dir).resolve()}")
    server = ServerWrapper(conf_dir)
    try:
        server.start()
    except (KeyboardInterrupt, SystemExit, ConnectionError) as e:
        logger.debug("Shutdown signal received.")
        # logger.exception(e)
    finally:
        logger.info("Shutting down scheduler.")
        server.shutdown()
        logger.debug("Scheduler shutdown complete.")
